#include "SDL_bmf.h"

#include "expat++.h"

#include <iostream>
#include <fstream>
#include <strstream>
#include <memory>

using namespace std;

// Construct a font object as in the example above.
SDL_BmapFont::SDL_BmapFont(const char* font,
		   Uint16 req_height,
		   Uint8 red,Uint8 grn,Uint8 blue) :
  fdesc_(font),
  font_bmap_(0),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(red),g_(grn),b_(blue),
  colorkey_b_((blue==0x00)?0xFF:0x00),
  bbmap_(0),
  no_color_(false)
{
  load_bmap_font();
}

// Construct a font object using an existing surface as
// the master bitmap. This can be used to allow multiple SDL_BmapFont
// objects with different colors and/or scaling factors to
// share a master bitmap.
SDL_BmapFont::SDL_BmapFont(SDL_Surface* master_bmap,
			   const char* font,
			   Uint16 req_height,
			   Uint8 red,Uint8 blue,Uint8 grn) :
  fdesc_(font),
  font_bmap_(master_bmap),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(red),g_(grn),b_(blue),
  colorkey_b_((blue==0x00)?0xFF:0x00),
  bbmap_(0),
  no_color_(false)
{
  load_bmap_font();
}

// Construct a font object using the given SDL_BmapFont as
// the bitmap. This can be used to allow multiple SDL_BmapFont
// objects with different colors and/or scaling factors to
// share a master bitmap.
SDL_BmapFont::SDL_BmapFont(const SDL_BmapFont& orig,
			   Uint16 req_height,
			   Uint8 red,Uint8 grn,Uint8 blue) :
  fdesc_(orig.fdesc_),
  font_bmap_(orig.font_bmap_),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(red),g_(grn),b_(blue),
  colorkey_b_((blue==0x00)?0xFF:0x00),
  bbmap_(0),
  no_color_(false)
{
  load_bmap_font();
}

// Construct a font object as in the example above.
SDL_BmapFont::SDL_BmapFont(const char* font,
		   Uint16 req_height,
		   SDL_Surface* bbmap) :
  fdesc_(font),
  font_bmap_(0),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(0),g_(0),b_(0),
  colorkey_b_(0),
  bbmap_(bbmap),
  no_color_(false)
{
  if (!bbmap_) no_color_ = true;
  load_bmap_font();
}

// Construct a font object using an existing surface as
// the master bitmap. This can be used to allow multiple SDL_BmapFont
// objects with different colors and/or scaling factors to
// share a master bitmap.
SDL_BmapFont::SDL_BmapFont(SDL_Surface* master_bmap,
			   const char* font,
			   Uint16 req_height,
			   SDL_Surface* bbmap) :
  fdesc_(font),
  font_bmap_(master_bmap),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(0),g_(0),b_(0),
  colorkey_b_(0),
  bbmap_(bbmap),
  no_color_(false)
{
  if (!bbmap_) no_color_ = true;
  load_bmap_font();
}

// Construct a font object using the given SDL_BmapFont as
// the bitmap. This can be used to allow multiple SDL_BmapFont
// objects with different colors and/or scaling factors to
// share a master bitmap.
SDL_BmapFont::SDL_BmapFont(const SDL_BmapFont& orig,
			   Uint16 req_height,
			   SDL_Surface* bbmap) :
  fdesc_(orig.fdesc_),
  font_bmap_(orig.font_bmap_),
  rel_font_bmap_(false),
  req_ht_(req_height),
  scale_(1),
  max_ht_(0),
  usable_(false),
  r_(0),g_(0),b_(0),
  colorkey_b_(0),
  bbmap_(bbmap),
  no_color_(false)
{
  if (!bbmap_) no_color_ = true;
  load_bmap_font();
}

SDL_BmapFont::~SDL_BmapFont()
{
  if (font_bmap_ && rel_font_bmap_) SDL_FreeSurface(font_bmap_);
  map<int,CharStruct*>::iterator it;
  for (it=char_bmaps_.begin(); it != char_bmaps_.end(); ++it) {
    CharStruct* cs = (*it).second;
    if (cs->bmap) SDL_FreeSurface(cs->bmap);
    delete cs;
  }
}

void SDL_BmapFont::load_bmap_font()
{
  // Try to parse the first argument as XML.
  XMLParser parser;
  XML_Error err;
  long errloc;
  istrstream iss(fdesc_.c_str(),fdesc_.size());
  const XMLTag* fdesc = 0;
  fdesc = parser.Parse(iss,err,errloc,errstr_);
  if (!fdesc) {

    //cout << "Could not parse " << fdesc_ << " -- " << errstr_ << endl;

    // If that fails, try to open it as a file and parse the
    // contents.
    ifstream ifs(fdesc_.c_str());
    if (ifs.good()) {
	  XMLParser parser;		// D.Finkler: new instance required
      fdesc = parser.Parse(ifs,err,errloc,errstr_);
    }
  }

  // If that fails, bail.
  if (!fdesc) {
    errstr_ = "Could not parse font description.";
    return;
  }

  //cout << "Got valid font description " << hex << fdesc << dec << endl;

  // fdesc is now the font description.
  if (loadXmlFont(fdesc)) {
    usable_ = true;
  } else {
    errstr_ += LXMLString(" - Could not load font description.");
  }
}


// D.Finkler:
// Render font and cache all chars in font
void SDL_BmapFont::CacheFont()
{
  map<int,CharStruct*>::iterator it;
  for (it=char_bmaps_.begin(); it != char_bmaps_.end(); ++it) {
    CharStruct* cs = (*it).second;
    scale_and_render_char(*cs);
  }
}

// D.Finkler:
// Render the given text and cache all chars in text
void SDL_BmapFont::CacheText(const char* text)
{
  while (*text != 0x00) {
    //cout << "Trying to blit " << (*text) << endl;
    CharStruct* cs = char_bmaps_[*(unsigned char*)text];
    if (cs) {
		scale_and_render_char(*cs);
    }
    // Cache char
    ++text;
  }
}

// Render the given text and return a transparent SDL_Surface
// containing the resulting bitmap.
SDL_Surface* SDL_BmapFont::DrawText(const char* text,int hgap)
{
  // Make a surface big enough to hold the rendered text.
  Uint16 ht, wd;
  GetTextDimension(text,&ht,&wd,hgap);
  SDL_Surface* tsurf = SDL_CreateRGBSurface(SDL_SWSURFACE,
					    wd,ht,
					    24,
					    0xFF0000,0x00FF00,0x0000FF,0);
  if (!tsurf) return 0;

  // Fill the surface with the colorkey value.
  //cout << "colorkey_b_ is " << hex << (int)colorkey_b_ << endl;
  SDL_FillRect(tsurf,0,SDL_MapRGB(tsurf->format,colorkey_b_,colorkey_b_,colorkey_b_));

  // Set the color key.
  SDL_SetColorKey(tsurf,SDL_SRCCOLORKEY,SDL_MapRGB(tsurf->format,colorkey_b_,colorkey_b_,colorkey_b_));

  // Render the text.
  DrawText(tsurf,0,0,text,hgap);

  // Display-format the surface.
  SDL_Surface* dsurf = SDL_DisplayFormat(tsurf);

  // Free the original surface.
  SDL_FreeSurface(tsurf);

  return dsurf;
}

// Render the given text directly onto the specified surface
// at the given coordinates.
void SDL_BmapFont::DrawText(SDL_Surface* dest,int x,int y,const char* text,int hgap)
{
  SDL_Rect srect;
  SDL_Rect drect;
  while (*text != 0x00) {
    //cout << "Trying to blit " << (*text) << endl;
    CharStruct* cs = char_bmaps_[*(unsigned char*)text];
    if (!cs) {
      //cout << "Could not find cstruct " << *text << endl;
      ++text;
      continue;
    }
    SDL_Surface* csurf = scale_and_render_char(*cs);
    if (!csurf) {
      //cout << "Could not render " << *text << endl;
      ++text;
      continue;
    }
    srect.x = 0;
    srect.y = 0;
    srect.w = csurf->w;
    srect.h = csurf->h;
    drect.x = x;
    drect.y = y+cs->desc;
    drect.w = csurf->w;
    drect.h = csurf->h;
    SDL_BlitSurface(csurf,&srect,dest,&drect);
    x += csurf->w;
    x += hgap;
    // No need to free csurf; it's the cached char.
    ++text;
  }
}

// Save the font as XML text. Return true on success, false
// on failure.
bool SDL_BmapFont::SaveAsXML(const char* xmlfile)
{
  ofstream of(xmlfile);
  if (!of.good()) {
    errstr_ = "Could not open XML file.";
    return false;
  }

  of << "<FONT type=\"encoded\">" << endl;
  map<int,CharStruct*>::iterator it;
  for (it=char_bmaps_.begin(); it != char_bmaps_.end(); ++it) {
    int code = (*it).first;
    CharStruct* cs = (*it).second;
    saveCharToStream(of,code,cs,false);
  }
  of << "</FONT>" << endl;

  return true;
}

// Save as a BMP file with auxiliary XML description.
bool SDL_BmapFont::SaveAsBMPAndXML(const char* bmapfile,const char* xmlfile)
{
  if (-1 == SDL_SaveBMP(font_bmap_,bmapfile)) {
    errstr_ = "Could not open bitmap file.";
    return false;
  }

  ofstream of(xmlfile);
  if (!of.good()) {
    errstr_ = "Could not open XML file.";
    return false;
  }

  of << "<FONT type=\"description\" bmapfile=\"" << bmapfile << "\">" << endl;
  map<int,CharStruct*>::iterator it;
  for (it=char_bmaps_.begin(); it != char_bmaps_.end(); ++it) {
    int code = (*it).first;
    CharStruct* cs = (*it).second;
    saveCharToStream(of,code,cs,true);
  }
  of << "</FONT>" << endl;

  return true;
}

// Get the maximum height of any character in the font.
int SDL_BmapFont::GetMaxHeight()
{
  return max_ht_;
}

// D.Finkler:
int SDL_BmapFont::GetCharWidth(const unsigned char ch)
{
    CharStruct* cs = char_bmaps_[ch];
    if (cs) 
		return (int)((float)cs->wd*scale_);
	return 0;
}

// Find the dimensions of the region necessary to hold the
// rendered string.
void SDL_BmapFont::GetTextDimension(const char* text,Uint16* ht,Uint16* wd,int hgap)
{
  *ht = 0;
  *wd = 0;
  bool firsttime = true;
  while (*text != 0x00) {
    CharStruct* cs = char_bmaps_[*(unsigned char*)text];
    if (!cs) {
      ++text;
      continue;
    }
    int cht = (int)((float)cs->ht*scale_) + cs->desc;
    if (cht > *ht) *ht = cht;
    *wd += (int)((float)cs->wd*scale_);
    if (firsttime) firsttime = false;
    else *wd += hgap;
    ++text;
  }
}

// Scale and render a particular character given its cache description.
SDL_Surface* SDL_BmapFont::scale_and_render_char(CharStruct& cs)
{
	// If we've already done it, return the scaled and rendered surface.
	if (cs.bmap) return cs.bmap;
	
	// (1) Make an SDL_Surface big enough to hold the scaled character.
	SDL_Surface* chsurf = SDL_CreateRGBSurface(SDL_SWSURFACE,
											   (int)((float)cs.wd*scale_),
											   (int)((float)cs.ht*scale_),
											   24,
											   0xFF0000,0x00FF00,0x0000FF,0);
	if (!chsurf) return 0;
	
	// Fill the entire surface black.
	//cout << "colorkey_b_ is " << hex << (int)colorkey_b_ << endl;
	SDL_FillRect(chsurf,0,SDL_MapRGB(chsurf->format,colorkey_b_,colorkey_b_,colorkey_b_));
	
	// Set black as the colorkey.
	if (0 != SDL_SetColorKey(chsurf,SDL_SRCCOLORKEY,SDL_MapRGB(chsurf->format,colorkey_b_,colorkey_b_,colorkey_b_))) {
		cout << "Error setting colorkey in character bitmap." << endl;
	}
	
	// (2) Move the pixel data for the character onto the surface, converting
	// all non-black pixels to the requested color.
	//cout << "Rendering char at " << cs.x << ',' << cs.y << " of size " << cs.wd << ',' << cs.ht << endl;
	int sw = (int)((float)cs.wd*scale_);
	int sh = (int)((float)cs.ht*scale_);
	for (int sy=0; sy<sh; ++sy) {
//		int my = (int)((float)cs.ht * ((float)sy/(float)sh));
		int my = (int)((float)cs.ht * (float)sy / (float)sh);
		for (int sx=0; sx<sw; ++sx) {
			Uint8 r=0,g=0,b=0;
			// Get the corresponding pixel in the master bmap.
//			int mx = (int)((float)cs.wd * ((float)sx/(float)sw));
			int mx = (int)((float)cs.wd * (float)sx / (float)sw);
			if (get_pixel(font_bmap_,cs.x+mx,cs.y+my,&r,&g,&b)) {
				if (no_color_) {
					// Just copy the pixels.
					set_pixel(chsurf,sx,sy,r,g,b);
				} else {
					// Need to do colorization.
					if (bbmap_) {
						// Use the corresponding pixel from the backing bmap.
						int bw=bbmap_->w;
						int bh=bbmap_->h;
						get_pixel(bbmap_,sx%bw,sy%bh,&r_,&g_,&b_);
					}
					// If no bbmap, r_ g_ b_ were set by the ctor.
					set_pixel(chsurf,sx,sy,r_,g_,b_);
				}
			}
		}
	}
	
	// (3) Displayformat the surface and install it in cs.
	cs.bmap = SDL_DisplayFormat(chsurf);
	SDL_FreeSurface(chsurf);
	
	// Now cs.bmap is a display-formatted, transparent bitmap of the
	// scaled character in the proper color.
	return cs.bmap;
}

// Get/set pixels in an SDL_Surface. get_pixel() returns true iff the
// selected pixel is not black.
void SDL_BmapFont::set_pixel(SDL_Surface* screen,int X,int Y,Uint8 r,Uint8 g,Uint8 b)
{
  Uint32   pixel;
  Uint8   *bits, bpp;

  pixel = SDL_MapRGB(screen->format, r, g, b);

  //if ( SDL_MUSTLOCK(screen) ) {
  // Will never happen; we only call set_pixel on SW surfaces.
  //  if ( SDL_LockSurface(screen) < 0 )
  //    return;
  //}

  /* Calculate the framebuffer offset of the screen coords */
  // Note: screen->pitch is the width of the screen in BYTES;
  // bpp is the width of a PIXEL in bytes.
  bpp = screen->format->BytesPerPixel;
  bits = ((Uint8 *)screen->pixels)+Y*screen->pitch+X*bpp;

  /* Set the pixel */
  switch(bpp) {
  case 1:
    *((Uint8 *)(bits)) = (Uint8)pixel;
    break;
  case 2:
    *((Uint16 *)(bits)) = (Uint16)pixel;
    break;
  case 3: { /* Format/endian independent */
    Uint8 r, g, b;

    r = (pixel>>screen->format->Rshift)&0xFF;
    g = (pixel>>screen->format->Gshift)&0xFF;
    b = (pixel>>screen->format->Bshift)&0xFF;
    *((bits)+screen->format->Rshift/8) = r; 
    *((bits)+screen->format->Gshift/8) = g;
    *((bits)+screen->format->Bshift/8) = b;
  }
  break;
  case 4:
    *((Uint32 *)(bits)) = (Uint32)pixel;
    break;
  }
}

bool SDL_BmapFont::get_pixel(SDL_Surface* screen,int X,int Y,Uint8* r,Uint8* g,Uint8* b)
{
	Uint32   pixel=0;
	Uint8   *bits, bpp;
	
	/* Calculate the framebuffer offset of the screen coords */
	bpp = screen->format->BytesPerPixel;
	bits = ((Uint8 *)screen->pixels)+Y*screen->pitch+X*bpp;
	
	/* Get the pixel */
	switch(bpp) {
		case 1:
			pixel = *((Uint8 *)(bits));
			////cout << "1bpp, raw pixel =  " << hex << pixel << dec << ' ';
			break;
		case 2:
			pixel = *((Uint16 *)(bits));
			////cout << "2bpp, raw pixel =  " << hex << pixel << dec << ' ';
			break;
		case 3: { /* Format/endian independent */
			////cout << "3bpp ";
			*r =
			*((bits)+screen->format->Rshift/8);
			*g = *((bits)+screen->format->Gshift/8);
			*b = *((bits)+screen->format->Bshift/8);
			////cout << "Getting pixel " << X << ',' << Y << ": rgb = " << hex << (int)*r << ',' << hex << (int)*g << ',' << hex << (int)*b << dec << endl;
			return ((*r>0x7F) || (*g>0x7F) ||(*b>0x7F));
		}
			break;
		case 4:
			////cout << "4bpp ";
			pixel = *((Uint32 *)(bits));
			break;
	}
	SDL_GetRGB(pixel,screen->format, r, g, b);
	////cout << "Getting pixel " << X << ',' << Y << ": rgb = " << hex << (int)*r << ',' << hex << (int)*g << ',' << hex << (int)*b << dec << endl;
	return ((*r>0x7F) || (*g>0x7F) ||(*b>0x7F));
}

// Convert an XML-encoded font into a bitmap. This produces the master font
// bmap for font_bmap_, and also extracts the cache descriptors into char_bmaps_.
// It either loads the master bitmap from a file, or synthesizes it from
// the per-character bitmap data contained in the XML font description.
bool SDL_BmapFont::loadXmlFont(const XMLTag* font)
{
  choose_scale_factor(font);
  //cout << "Scale factor is " << scale_ << endl;

  XMLString type;
  if (font->AttrValue(XML_FONT_TYPE_ARG,&type) == 0) {
    if (type == XML_FONT_CONSTANT) return loadConstantSzFont(font);
    if (type == XML_FONT_DESCRIBED) return loadXMLDescribedFont(font);
    if (type == XML_FONT_ENCODED) return loadXMLEncodedFont(font);
  }

  errstr_ += "-- Unknown font type.";
  return false;
}

bool SDL_BmapFont::loadConstantSzFont(const XMLTag* font)
{
  long ht=0,wd=0,rows=0,cols=0;
  if (font->AttrValue(XML_WIDTH_ARG,&wd) != 0) {
    errstr_ = "No width specified for fixed-size font.";
    return false;
  }
  if (font->AttrValue(XML_HEIGHT_ARG,&ht) != 0) {
    errstr_ = "No height specified for fixed-size font.";
    return false;
  }
  if (font->AttrValue(XML_ROWS_ARG,&rows) != 0) {
    errstr_ = "No array row count specified for fixed-size font.";
    return false;
  }
  if (font->AttrValue(XML_COLUMNS_ARG,&cols) != 0) {
    errstr_ = "No array column count specified for fixed-size font.";
    return false;
  }

  //cout << "w=" << wd << ", ht=" << ht << ", rows=" << rows << ", cols=" << cols << endl;

  if (!font_bmap_) {
    bool rc = tryToLoadBmapFile(font);
    if (!rc) {
      errstr_ += "-- Could not load bitmap file.";
      return false;
    }
  }

  gen_const_descs(wd,ht,rows,cols);
}

bool SDL_BmapFont::loadXMLDescribedFont(const XMLTag* font)
{
  if (!extract_char_descs(font,false)) {
    errstr_ = "Could not load all character-descriptors for XML-described font.";
    return false;
  }
  bool rc = true;
  if (!font_bmap_) rc = tryToLoadBmapFile(font);
  return rc;
}

bool SDL_BmapFont::loadXMLEncodedFont(const XMLTag* font)
{
  //cout << "Loading XML-encoded font..." << endl;
  if (!extract_char_descs(font,true)) {
    errstr_ = "Could not load all character-descriptors for XML-encoded font.";
    return false;
  }
}

// Utilities to operate on the XML font description.
int SDL_BmapFont::fontTotalWidth(const XMLTag* font)
{
  //cout << "Finding font total width..." << endl;
  // Three possibilities:
  // (1) It's a constant-size font.
  long total_wd = 0,wd=0,rows=0,cols=0;
  if (font->NElements() == 0) {
    if (font->AttrValue(XML_WIDTH_ARG,&wd) != 0) {
      return 0;
    }
    if (font->AttrValue(XML_ROWS_ARG,&rows) != 0) {
      return 0;
    }
    if (font->AttrValue(XML_COLUMNS_ARG,&cols) != 0) {
      return 0;
    }
    return rows*cols*wd;
  }

  // (2) It's a variable-size font. We must get the max
  // height of any character.
  int nel = font->NElements();
  for (int e=0; e<nel; ++e) {
    const XMLElement* el = font->NthElement(e);
    if (el->GetType() == XMLElement::TEXT) continue;
    const XMLTag* tag = (XMLTag*)el;
    if (tag->GetName() != XML_CHARACTER_TAG) continue;
    long wd = 0;
    // (3) It's an ill-formed font description; bail.
    if (tag->AttrValue(XML_WIDTH_ARG,&wd) != 0) return 0;
    total_wd += wd;
  }

  //cout << "Font total width is " << total_wd << endl;
  return total_wd;
}

int SDL_BmapFont::fontMaxHeight(const XMLTag* font)
{
  //cout << "Finding font max height, font = " << hex << font << dec << "..." << endl; 
  // Three possibilities:
  // (1) It's a constant-size font.
  long ht = 0;
  if (font->NElements() == 0) {
    //cout << "Constant-size font." << endl;
    if (font->AttrValue(XML_HEIGHT_ARG,&ht) == 0) {
      return ht;
    } else return 0;
  }
  //cout << "Variable-size font." << endl;

  // (2) It's a variable-size font. We must get the max
  // height of any character.
  int nel = font->NElements();
  long max_ht = 0;
  for (int e=0; e<nel; ++e) {
    const XMLElement* el = font->NthElement(e);
    if (el->GetType() == XMLElement::TEXT) continue;
    const XMLTag* tag = (XMLTag*)el;
    if (tag->GetName() != XML_CHARACTER_TAG) continue;
    long ht = 0;
    // (3) It's an ill-formed font description; bail.
    if (tag->AttrValue(XML_HEIGHT_ARG,&ht) != 0) return 0;
    //long desc = 0;
    // Descenders increase the effective character height.
    // (But we don't care here!)
    //tag->AttrValue(XML_DESCENDER_ARG,&desc);
    //ht += desc;
    if (ht>max_ht) max_ht=ht;
  }
  return max_ht;
}

SDL_BmapFont::XBmap::XBmap(const XMLTag* ch) : bits(""),wd(0),ht(0),sz(0)
{
  //cout << "Making expanded bitmap..." << endl;

  bits = "";

  // Get the width and height.
  if (ch->AttrValue(XML_WIDTH_ARG,&wd) != 0) {
    return;
  }
  if (ch->AttrValue(XML_HEIGHT_ARG,&ht) != 0) {
    return;
  }
  sz = wd*ht;

  //cout << "w=" << wd << ", h=" << ht << endl;

  const XMLElement* el = ch->NthElement(0);
  if (!el) return;
  if (el->GetType() != XMLElement::TEXT) {
    //cout << "Non-text element in CH tag." << endl;
    return;
  }
  XMLString tx = ((XMLText*)el)->GetText();
  for (int ii=0; ii<tx.length(); ++ii) {
    XML_Char cc = tx[ii];
    if ((cc == XML_BMAP_SET) || (cc == XML_BMAP_CLR)) {
      bits += cc;
    }
  }
  while (bits.length() < sz) bits += XML_BMAP_CLR;
}

// Move an expanded bitmap to an SDL_Surface.
void SDL_BmapFont::exBmapToSurface(XBmap& xbmap,SDL_Surface* img,int x,int y)
{
  //cout << "Writing expanded bmap on surface..." << endl;
  if (x+xbmap.getWd() > img->w) {
    //cout << "Width violation - x=" << x << ", img->w=" << img->w << ", width=" << xbmap.getWd() << endl;
    return;
  }
  if (y+xbmap.getHt() > img->h) {
    //cout << "Height violation - y=" << y << ", img->h=" << img->h << ", height=" << xbmap.getHt() << endl;
    return;
  }
  for (int yy=0; yy<xbmap.getHt(); ++yy) {
    for (int xx=0; xx<xbmap.getWd(); ++xx) {
      if (xbmap.getPixel(xx,yy)) {
	set_pixel(img,x+xx,y+yy,0xFF,0xFF,0xFF);
	//cout << '*';
      } else {
	set_pixel(img,x+xx,y+yy,0x00,0x00,0x00);
	//cout << '-';
      }
    }
    //cout << endl;
  }
}

bool SDL_BmapFont::tryToLoadBmapFile(const XMLTag* font)
{
  XMLString fname;
  if (font->AttrValue(XML_FILE_ARG,&fname) != 0) {
    errstr_ = "No bmapfile name for XML-described font.";
    return false;
  }
  SDL_Surface* bmap = SDL_LoadBMP(fname.c_str());
  if (bmap) rel_font_bmap_ = true;
  else {
    errstr_ = LXMLString("Could not load bmapfile for XML-described font.");
	return false;
  }
  font_bmap_ = bmap;

  //cout << "font_bmap_ wd,ht = " << font_bmap_->w << ',' << font_bmap_->h << endl;

  return true;
}

void SDL_BmapFont::choose_scale_factor(const XMLTag* font)
{
  int max_ht = fontMaxHeight(font);

  //if (max_ht == 0) cout << "Could not compute max height." << endl;

  scale_ = (float)req_ht_/(float)max_ht;
  max_ht_ = (int)((float)max_ht * scale_);
}

bool SDL_BmapFont::extract_char_descs(const XMLTag* font,bool make_bmap)
{
  //cout << "Extracting char descriptions..." << endl;
  if (make_bmap) {
    int ht = fontMaxHeight(font);
    int wd = fontTotalWidth(font);
    if (!ht || !wd) {
      errstr_ = "Either height or width of font is 0.";
      return false;
    }
    font_bmap_ = SDL_CreateRGBSurface(SDL_SWSURFACE,wd,ht,8,0,0,0,0);
    if (!font_bmap_) {
      errstr_ = "Could not create in-memory font bitmap.";
      return false;
    }

    // Set up the palette. We only need black and white.
    SDL_Color clrs[2];
    clrs[0].r=0; clrs[0].g=0; clrs[0].b=0;
    clrs[1].r=0xFF; clrs[1].g=0xFF; clrs[1].b=0xFF;
    if (SDL_SetColors(font_bmap_,clrs,0,2) != 1) {
      errstr_ = "Could not set palette on in-memory bitmap.";
      return false; // Seems unlikely...
    }

    SDL_FillRect(font_bmap_,0,SDL_MapRGB(font_bmap_->format,0,0,0));
    rel_font_bmap_ = true;
  }

  int atx = 0;
  int nel = font->NElements();
  for (int elidx=0; elidx<nel; ++elidx) {
    const XMLElement* el = font->NthElement(elidx);
    if (el->GetType() == XMLElement::TAG) {
      const XMLTag* ch = (XMLTag*)el;
      if (ch->GetName() != XML_CHARACTER_TAG) continue;
      if (!extract_one_desc(ch,atx,make_bmap)) return false;
    }
  }
}

void SDL_BmapFont::gen_const_descs(int wd,int ht,int rows,int cols)
{
  int code=0;
  for (int rr=0; rr<rows; ++rr) {
    for (int cc=0; cc<cols; ++cc) {
      int x = cc*wd;
      int y = rr*ht;
      int descend = 0;
      CharStruct* cs = new CharStruct(x,y,wd,ht,descend);
      char_bmaps_[code] = cs;
      //cout << "Installed cstruct for " << code << '(' << (char)code << ')' << " x,y = " << x << ',' << y << "; w,h = " << wd << ',' << ht << " row,col = " << rr << ',' << cc << endl;
      ++code;
    }
  }
}

bool SDL_BmapFont::extract_one_desc(const XMLTag* ch,int& atx,bool make_bmap)
{
  //cout << "Extracting description..." << endl;
  long wd=0,ht=0,code=0,descend=0;
  long x=0,y=0;

  // We should always find a code, height and width.
  if (ch->AttrValue(XML_CHARCODE_ARG,&code) != XMLTag::ATTR_SUCCESS) {
    errstr_ = "Missing character code in CH tag.";
    return false;
  }
  //cout << "Description for " << code << endl;
  if (ch->AttrValue(XML_HEIGHT_ARG,&ht) != XMLTag::ATTR_SUCCESS) {
    errstr_ = "Missing character height in CH tag.";
    return false;
  }
  if (ch->AttrValue(XML_WIDTH_ARG,&wd) != XMLTag::ATTR_SUCCESS) {
    errstr_ = "Missing character width in CH tag.";
    return false;
  }
  ch->AttrValue(XML_DESCENDER_ARG,&descend);

  // Groovy. Now. if we're not building the master bitmap ourselves, we
  // also need the bitmap x and y coords.
  if (!make_bmap) {
    if (ch->AttrValue(XML_XCOORD_ARG,&x) != XMLTag::ATTR_SUCCESS) {
      errstr_ = "Missing character x coord in CH tag.";
      return false;
    }
    if (ch->AttrValue(XML_YCOORD_ARG,&y) != XMLTag::ATTR_SUCCESS) {
      errstr_ = "Missing character y coord in CH tag.";
      return false;
    }
  } else {
    x = atx;
    y=0;
  }

  // Excellent. Build the descriptor and install it in the character map.
  CharStruct* cs = new CharStruct(x,y,wd,ht,(Uint16)(descend*scale_));
  char_bmaps_[code] = cs;

  // If we need to build the bitmap, do so.
  if (make_bmap) {
    XBmap xbm(ch);
    exBmapToSurface(xbm,font_bmap_,x,y);
    atx += wd;
  }

  return true;
}

ostream& SDL_BmapFont::saveCharToStream(ostream& os,int code,CharStruct* cs,bool desc_only)
{
  Uint8 r,g,b;
  os << "<CH c=\"" << code << "\" w=\"" << cs->wd << "\" h=\"" << cs->ht << "\" "
     << "d=\"" << (int)cs->desc/scale_ << "\" ";
  if (desc_only) {
    os << "x=\"" << cs->x << "\" y=\"" << cs->y << "\"/>" << endl;
  } else {
    os << ">" << endl;
    for (int y=cs->y; y<cs->y+cs->ht; ++y) {
      for (int x=cs->x; x<cs->x+cs->wd; ++x) {
	if (get_pixel(font_bmap_,x,y,&r,&g,&b)) {
	  os << "*";
	} else {
	  os << "-";
	}
      }
      os << endl;
    }
    os << "</CH>" << endl;
  }
}
