// Copyright (c) 2000 Joseph A. Knapka.
//
// License: GPL/LGPL, as appropriate.
#include "SDL_bmf.h"

#include <algorithm>
#include <iostream>
#include <strstream>

using namespace std;

void waitForKey(SDL_Surface* screen,SDL_Surface* wait_bmap)
{
  SDL_Delay(2000);

  int xpos = screen->w/2 - wait_bmap->w/2;
  int ypos = screen->h/2 - wait_bmap->h/2;
  SDL_Rect srect,drect;
  srect.x=0;srect.y=0;srect.w=wait_bmap->w;srect.h=wait_bmap->h;
  drect.x=xpos;drect.y=ypos;drect.w=wait_bmap->w;drect.h=wait_bmap->h;
  SDL_BlitSurface(wait_bmap,&srect,screen,&drect);
  SDL_Flip(screen);

  SDL_Event ev;
  while (1) {
    SDL_Delay(100);
    SDL_PumpEvents();
    SDL_PollEvent(&ev);
    if (ev.type == SDL_KEYDOWN) return;
  }
}

int main(int argc,char* argv[])
{

  // Initialize SDL: 640x480, 8bpp.
  SDL_Init(SDL_INIT_VIDEO);
  SDL_Surface* screen = SDL_SetVideoMode(640,480,8,SDL_SWSURFACE);
  SDL_Surface* wait_bmap = 0;

  atexit(SDL_Quit);

  // Set palette for black, white, red, green, blue, and yellow.
  SDL_Color colors[8] = {
    {0x00,0x00,0x00,0},
    {0xFF,0x00,0x00,0},
    {0xFF,0xFF,0x00,0},
    {0x00,0xFF,0x00,0},
    {0x00,0x00,0xFF,0},
    {0xFF,0xFF,0xFF,0},
    {0xFF,0x00,0xFF,0},
    {0x00,0xFF,0xFF,0},
  };
  SDL_SetColors(screen,colors,0,8);

  // Fill the screen with blue.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0xFF));

  SDL_Flip(screen);

  // Load the "LargeFont" font from the SDL_Console distro. Use
  // size 18 (the natural font size) and 36 (scale *2), and colors
  // red and blue.
  SDL_BmapFont lgred18("<FONT type=\"constant\" h=\"18\" w=\"16\" rows=\"1\" columns=\"256\" bmapfile=\"LargeFont.bmp\"/>",
		       18,
		       0xFF,0x00,0x00);
  SDL_BmapFont lgblue36("<FONT type=\"constant\" h=\"18\" w=\"16\" rows=\"1\" columns=\"256\" bmapfile=\"LargeFont.bmp\"/>",
			36,
			0x00,0x00,0xFF);
  SDL_BmapFont lgppl9("<FONT type=\"constant\" h=\"18\" w=\"16\" rows=\"1\" columns=\"256\" bmapfile=\"LargeFont.bmp\"/>",
		       9,
		       0xFF,0x00,0xFF);

  // Load the "Proportional8" font from an XML-encoded font file. Use
  // sizes 11, 22, 33, and 44, and colors red, green, blue, and yellow.
  cout << "green8" << endl;
  SDL_BmapFont prgreen8( "Proportional8.xml", 8,0x00,0xFF,0x00);
  SDL_BmapFont prred16(prgreen8,16,0xFF,0x00,0x00);
  SDL_BmapFont prblue24(prgreen8,24,0x00,0x00,0xFF);
  SDL_BmapFont prylw32(prgreen8,32,0xFF,0xFF,0x00);
  SDL_BmapFont prred48(lgppl9,40,0xFF,0x00,0x00);

  // Write some text here and there.
  if (lgred18.Usable()) {
    lgred18.DrawText(screen,10,10,"ABCDEFGHIJKLMnopqrstuvwxyz");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (red): " << lgred18.GetErrorDesc() << endl;
  }

  if (lgblue36.Usable()) {
    lgblue36.DrawText(screen,10,50,"Hello, world!");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (blue): " << lgblue36.GetErrorDesc() << endl;
  }

  if (lgppl9.Usable()) {
    lgppl9.DrawText(screen,100,300,"ABCDEFGHIJKLMnopqrstuvwxyz");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (purple): " << lgppl9.GetErrorDesc() << endl;
  }

  if (prgreen8.Usable()) {
    // Note: this font has internal padding, so no hgap necessary.
    prgreen8.DrawText(screen,10,100,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (green): " << prgreen8.GetErrorDesc() << endl;
  }
  if (prred16.Usable()) {
    prred16.DrawText(screen,10,116,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
    prred16.DrawText(screen,10,400,"ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    SDL_Flip(screen);
    prred16.DrawText(screen,10,420,"abcdefghijklmnopqrstuvwxyz");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (red): " << prred16.GetErrorDesc() << endl;
  }
  if (prblue24.Usable()) {
    prblue24.DrawText(screen,10,140,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (blue): " << prblue24.GetErrorDesc() << endl;
  }
  if (prylw32.Usable()) {
    prylw32.DrawText(screen,10,180,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (yellow): " << prylw32.GetErrorDesc() << endl;
    exit(1);
  }
  if (prred48.Usable()) {
    wait_bmap = prred48.DrawText("Press Any Key...");
  } else {
    cout << "Prop8 (red): " << prred48.GetErrorDesc() << endl;
    exit(1);
  }

  SDL_Flip(screen);

  waitForKey(screen,wait_bmap);

  // Fill the screen with black.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0x00));
  SDL_Flip(screen);

  // Write some text here and there.
  if (lgred18.Usable()) {
    lgred18.DrawText(screen,10,10,"ABCDEFGHIJKLMnopqrstuvwxyz");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (red): " << lgred18.GetErrorDesc() << endl;
  }

  if (lgblue36.Usable()) {
    lgblue36.DrawText(screen,10,50,"Hello, world!");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (blue): " << lgblue36.GetErrorDesc() << endl;
  }

  if (lgppl9.Usable()) {
    lgppl9.DrawText(screen,100,300,"Hello, world!");
    SDL_Flip(screen);
  } else {
    cout << "LargeFont (purple): " << lgppl9.GetErrorDesc() << endl;
  }

  if (prgreen8.Usable()) {
    // Note: this font has internal padding, so no hgap necessary.
    prgreen8.DrawText(screen,10,100,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (green): " << prgreen8.GetErrorDesc() << endl;
  }
  if (prred16.Usable()) {
    prred16.DrawText(screen,10,116,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
    prred16.DrawText(screen,10,400,"ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    SDL_Flip(screen);
    prred16.DrawText(screen,10,420,"abcdefghijklmnopqrstuvwxyz");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (red): " << prred16.GetErrorDesc() << endl;
  }
  if (prblue24.Usable()) {
    prblue24.DrawText(screen,10,140,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (blue): " << prblue24.GetErrorDesc() << endl;
  }
  if (prylw32.Usable()) {
    prylw32.DrawText(screen,10,180,"The quick brown fox jumps over the lazy dog.");
    SDL_Flip(screen);
  } else {
    cout << "Prop8 (yellow): " << prylw32.GetErrorDesc() << endl;
  }

  waitForKey(screen,wait_bmap);

  // Fill the screen with black.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0x00));
  SDL_Flip(screen);

  int y=10;
  for (int psz=4; psz<=40; psz+=2) {
    SDL_BmapFont lgvar(lgred18,psz,0x00,0xFF,0xFF);
    if (lgvar.Usable()) {
      lgvar.DrawText(screen,10,y,"Scaled 4 to 40 pixels");
    } else {
      cout << "Large (variable): " << lgvar.GetErrorDesc() << endl;
    }
    y += psz+2;
  }

  SDL_Flip(screen);
  waitForKey(screen,wait_bmap);

  // Fill the screen with black.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0x00));
  SDL_Flip(screen);

  SDL_Surface* bbmap = SDL_LoadBMP("bbmap.bmp");
  if (!bbmap) {
    cout << "Could not load backing bitmap." << endl;
    exit(1);
  }
  y=10;
  for (int psz=4; psz<=40; psz+=2) {
    SDL_BmapFont lgvar(lgred18,psz,bbmap);
    if (lgvar.Usable()) {
      lgvar.DrawText(screen,10,y,"A backing bitmap.");
    } else {
      cout << "Large (variable): " << lgvar.GetErrorDesc() << endl;
    }
    y += psz+2;
  }

  SDL_Flip(screen);
  waitForKey(screen,wait_bmap);

  // Fill the screen with black.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0x00));
  SDL_Flip(screen);

  y=10;
  for (int psz=4; psz<=40; psz+=2) {
    SDL_BmapFont lg_nocolor(lgred18,psz,(SDL_Surface*)0);
    if (lg_nocolor.Usable()) {
      lg_nocolor.DrawText(screen,10,y,"Raw Font Bitmap");
    } else {
      cout << "Large (variable): " << lg_nocolor.GetErrorDesc() << endl;
    }
    y += psz+2;
  }

  SDL_Flip(screen);
  waitForKey(screen,wait_bmap);

  // Fill the screen with black.
  SDL_FillRect(screen,0,SDL_MapRGB(screen->format,0x00,0x00,0x00));
  SDL_Flip(screen);

  SDL_Rect drect;
  drect.x=0;
  drect.y=200;
  drect.w=640;
  drect.h=205;
  SDL_BmapFont lgb200(lgred18,200,bbmap);
  SDL_Event ev;
  for (int ii=0; ii<1000; ++ii) {
    SDL_FillRect(screen,&drect,SDL_MapRGB(screen->format,0x00,0x00,0x00));
    char count[16];
    std::fill(count,count+16,0x00);
    ostrstream os(count,16);
    os << ii;
    lgb200.DrawText(screen,50,200,os.str());
    SDL_Flip(screen);
    if (!(ii%10)) {
      SDL_PumpEvents();
      if (SDL_PollEvent(&ev) != 0) {
	if (ev.type == SDL_KEYDOWN) break;
      }
    }
  }

  cout << "Saving fonts..." << endl;
  if (!lgred18.SaveAsXML("LargeFont.xml")) {
    cout << "Could not save Large - " << lgred18.GetErrorDesc() << endl;
  }
  if (!prgreen8.SaveAsBMPAndXML("Prop8.bmp","Prop8.xml")) {
    cout << "Could not save Proportional - " << prgreen8.GetErrorDesc() << endl;
  }

  SDL_Quit();

  return 0;
}
