// Bitmapped font <--> XML conversion tool.
//
// Copyright (c) 2000 Joseph A. Knapka.
//
// License: GPL/LGPL, as appropriate.

#include <iostream>
#include <fstream>
#include <string>
using namespace std;

#include "SDL_bmf.h"

void usage()
{
  cout << "Try: fontxml -toxml <descfile> <xmlfile>" << endl;
  cout << "     fontxml -tobmp <xmlfile> <descfile> <bmapfile>" << endl;
  cout << "Where <descfile> is an XML font-description file, <xmlfile> is a" << endl;
  cout << "file containing a complete XML-encoded font, and <bmapfile> is a" << endl;
  cout << "Windows BMP-format file containing a font bitmap." << endl;
  exit(0);
}

void convertToXml(const char* xmldesc,const char* xmlout)
{
  SDL_BmapFont font(xmldesc,16,0,0,0);
  if (!font.Usable()) {
    cerr << "Could not load font: " << font.GetErrorDesc() << endl;
    exit(1);
  }

  if (!font.SaveAsXML(xmlout)) {
    cerr << "Could not save font: " << font.GetErrorDesc() << endl;
    exit(1);
  }
}

void convertToXmlBmp(const char* xmlfont,const char* xmldesc,const char* bmapfile)
{
  SDL_BmapFont font(xmlfont,16,0,0,0);
  if (!font.Usable()) {
    cerr << "Could not load font: " << font.GetErrorDesc() << endl;
    exit(1);
  }

  if (!font.SaveAsBMPAndXML(bmapfile,xmldesc)) {
    cerr << "Could not save font: " << font.GetErrorDesc() << endl;
    exit(1);
  }
}

int main(int argc,char* argv[])
{
  if (argc<3) {
    usage();
  }

  SDL_Init(SDL_INIT_VIDEO);
  atexit(SDL_Quit);

  if (string(argv[1]) == "-toxml") {
    if (argc<4) usage();
    convertToXml(argv[2],argv[3]);
  }
  if (string(argv[1]) == "-tobmp") {
    if (argc<5) usage();
    convertToXmlBmp(argv[2],argv[3],argv[4]);
  }

}
