#include "expat++.h"
#include <iostream>
#include <fstream>
#include <memory>
using namespace std;

int ind = -1;

void indent()
{
  for (int i=0; i<ind; ++i) cout << "  ";
}

void print_text(const XMLText* text)
{
  basic_string<XML_Char> tx = text->GetText();
  indent();
  for (int ii=0; ii<tx.length(); ++ii) {
    XML_Char cc = tx[ii];
    if (cc == '\n') {
      cout << endl;
      indent();
    } else {
      cout << cc;
    }
  }
  cout << endl;
}

void print_tag(XMLTag* tag)
{
  // Pretty-print the XML.
  ++ind;

  const XMLTag* found = tag->FindMatchingSubtag("<ANNZE piglet=\"yes!\" booger=\"no!\"/>");
  if (found) { cout << "YAY, this tag has an Annze-piglet!" << endl; }

  found = tag->FindMatchingSubtag("<ANNZE booger=\"yes!\"/>");
  if (found) { cout << "BOGUE, this tag has an Annze-booger!" << endl; }

  indent();
  cout << '<' << tag->GetName();
  vector<basic_string<XML_Char> > attrs = tag->AttrNames();
  for (int aa=0; aa< attrs.size(); ++aa) {
    basic_string<XML_Char> val;
    tag->AttrValue(attrs[aa],&val);
    cout << ' ' << attrs[aa] << "=\"" << val << '\"';
  }
  if (tag->NElements() == 0) {
    cout << "/>" << endl;
  } else {
    cout << '>' << endl;
    for (int el=0; el<tag->NElements(); ++el) {
      const XMLElement* xel = tag->NthElement(el);
      if (xel->GetType() == XMLElement::TEXT) {
	++ind;
	print_text((const XMLText*)xel);
	--ind;
      } else {
	print_tag((XMLTag*)xel);
      }
    }
    indent();
    cout << "</" << tag->GetName() << '>' << endl;
  }
  --ind;
}

int main(int argc,char* argv[])
{
  if (argc<2) {
    cout << "Try xmltest <filename>" << endl;
    exit(0);
  }

  ifstream in(argv[1]);
  XML_Error err;
  long errline;
  long errcol;
  basic_string<XML_Char> errstr;
  XMLParser parser;
  auto_ptr<XMLTag> tag(parser.Parse(in,err,errline,errcol,errstr));
  if (!tag.get()) {
    cout << "Parse error at line " << errline << ", column " << errcol
	 << ": " << errstr << endl;
    exit(1);
  }

  print_tag(tag.get());

  return 0;
}
